'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _mapValues = require('lodash/mapValues');

var _mapValues2 = _interopRequireDefault(_mapValues);

var _constant = require('lodash/constant');

var _constant2 = _interopRequireDefault(_constant);

var _cloneDeep = require('lodash/cloneDeep');

var _cloneDeep2 = _interopRequireDefault(_cloneDeep);

var _sum = require('lodash/sum');

var _sum2 = _interopRequireDefault(_sum);

var _values = require('lodash/values');

var _values2 = _interopRequireDefault(_values);

var _keys = require('lodash/keys');

var _keys2 = _interopRequireDefault(_keys);

var _reduce = require('lodash/reduce');

var _reduce2 = _interopRequireDefault(_reduce);

var _forEach = require('lodash/forEach');

var _forEach2 = _interopRequireDefault(_forEach);

var _isNumber = require('lodash/isNumber');

var _isNumber2 = _interopRequireDefault(_isNumber);

var _clamp = require('lodash/clamp');

var _clamp2 = _interopRequireDefault(_clamp);

var _has = require('lodash/has');

var _has2 = _interopRequireDefault(_has);

var _isUndefined = require('lodash/isUndefined');

var _isUndefined2 = _interopRequireDefault(_isUndefined);

var _isObject = require('lodash/isObject');

var _isObject2 = _interopRequireDefault(_isObject);

var _isArray = require('lodash/isArray');

var _isArray2 = _interopRequireDefault(_isArray);

var _includes = require('lodash/includes');

var _includes2 = _interopRequireDefault(_includes);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var _ = { mapValues: _mapValues2.default, forEach: _forEach2.default, constant: _constant2.default, cloneDeep: _cloneDeep2.default,
  sum: _sum2.default, values: _values2.default, keys: _keys2.default, reduce: _reduce2.default, isNumber: _isNumber2.default, clamp: _clamp2.default, has: _has2.default,
  isUndefined: _isUndefined2.default, isObject: _isObject2.default, isArray: _isArray2.default, includes: _includes2.default

  /**
   * const roller = new Brng(config)
   *
   * Constructor parameters:
   *  originalProportions [REQUIRED] {Object[String/Number:Number]} -- key-value mapping of
   *    weighted proportions. For example {mickeyd: 3, jackinthebox: 3, burgerking: 2, whataburger: 10}
   *  config {Object}
   *  config.random {Function} -- function that returns random number 0 - 1. Defaults to Math.random
   *  config.keepHistory {Boolean} -- if true, keep the roll history
   *  config.bias {Number} -- between 0 and 4. The higher the bias, the more it
   *    favors values less chosen. If 0, it's basically a normal RNG. Defaults to 1.
   *  config.repeatTolerance {Number} -- between 0 and 1. The lower the tolerance, the more
   *    likely Brng will re-roll if it's a repeat. If 0, it'll never repeat. Defaults to 1.
   *
   * Public methods:
   *  roll() -- selects a random value; remembers previous rolls.
   *  roll(value) -- force select the value from your original proportions. Ignores all criteria.
   *  roll({exclude: [value1, value2]}) -- select a value that excludes any values in the array
   * 
   *  flip(), pick(), select(), choose(), randomize() -- aliases of `roll()`
   *  reset() -- resets all history and resets previous rolls
   *  undo () -- undo the previous roll. must have `config.keepHistory === true`
   *
   * Public readable values (if keepHistory === true):
   *  historyArray -- array of every single previous roll
   *  historyMapping -- hash of key to number of times rolled in total
   *
   * Public static values for common default proportions:
   *  Brng.defaultProportions.one6SidedDie -- proportions for rolling 1 6-sided die
   *  Brng.defaultProportions.two6SidedDice -- proportions for rolling 2 6-sided dice
   *  Brng.defaultProportions.coin -- proportions for flipping 1 coin
   */
};
var Brng = function () {
  function Brng(originalProportions) {
    var _this = this;

    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    _classCallCheck(this, Brng);

    this.flip = this.roll;
    this.pick = this.roll;
    this.select = this.roll;
    this.choose = this.roll;
    this.randomize = this.roll;

    this.keepHistory = !!config.keepHistory;
    if (this.keepHistory) {
      this.historyMapping = _.mapValues(originalProportions, _.constant(0));
      this.historyArray = [];
    }

    this.previousRoll = null;
    this.repeatTolerance = _.isNumber(config.repeatTolerance) ? _.clamp(config.repeatTolerance, 0, 1) : 1;
    this.random = config.random || Math.random;
    this.originalProportions = _.cloneDeep(originalProportions);
    this.proportions = _.cloneDeep(originalProportions);

    var sumTotal = _.sum(_.values(originalProportions));
    this.originalProbabilities = _.mapValues(originalProportions, function (number) {
      return number / sumTotal;
    });
    this.possibleKeys = _.keys(originalProportions);

    var baseValueToRedistribute = _.reduce(originalProportions, function (sum, proportionValue, keyId) {
      return sum + proportionValue * _this.originalProbabilities[keyId];
    }, 0);
    var biasMultiplier = _.isNumber(config.bias) ? _.clamp(config.bias, 0, 4) : 1;
    this.valueToRedistribute = baseValueToRedistribute * biasMultiplier;
  }

  // The first step of the algorithm: given a weighted proportion map, randomly select a value


  _createClass(Brng, [{
    key: 'chooseKeyFromProportions',
    value: function chooseKeyFromProportions() {
      var sumTotal = _.sum(_.values(this.proportions));

      var rawRoll = this.random() * sumTotal;
      var currentSum = 0;

      for (var i = 0; i < this.possibleKeys.length; i++) {
        var chosenKey = this.possibleKeys[i];
        var proportionValue = this.proportions[chosenKey];
        currentSum += proportionValue;
        if (currentSum > rawRoll) {
          return chosenKey;
        }
      }

      console.warn('Code should not reach here.');
    }

    // The second step of the algorithm: given the selected value, shift the proportion map around that value
    // WARNING: method writes to this.proportions

  }, {
    key: 'shiftProportions',
    value: function shiftProportions(keyChosen) {
      var _this2 = this;

      this.proportions[keyChosen] = this.proportions[keyChosen] - this.valueToRedistribute;
      _.forEach(this.possibleKeys, function (otherKey) {
        _this2.proportions[otherKey] = _this2.proportions[otherKey] + _this2.valueToRedistribute * _this2.originalProbabilities[otherKey];
      });
    }

    // the reverse of shiftProportions() for undo
    // WARNING: method writes to this.proportions

  }, {
    key: 'reverseShiftProportions',
    value: function reverseShiftProportions(keyChosen) {
      var _this3 = this;

      this.proportions[keyChosen] = this.proportions[keyChosen] + this.valueToRedistribute;
      _.forEach(this.possibleKeys, function (otherKey) {
        _this3.proportions[otherKey] = _this3.proportions[otherKey] - _this3.valueToRedistribute * _this3.originalProbabilities[otherKey];
      });
    }
  }, {
    key: 'passCriteria',
    value: function passCriteria(keyChosen) {
      if (this.previousRoll === keyChosen && this.repeatTolerance < 1) {
        return this.random() < this.repeatTolerance;
      } else {
        return true;
      }
    }
  }, {
    key: 'rollSideEffects',
    value: function rollSideEffects(keyChosen) {
      this.previousRoll = keyChosen;
      if (this.keepHistory) {
        this.historyMapping[keyChosen] = this.historyMapping[keyChosen] + 1;
        this.historyArray.push(keyChosen);
      }
      return keyChosen;
    }
  }, {
    key: 'selectKeyAtRandom',
    value: function selectKeyAtRandom(config) {
      var excludeList = [];
      var keyChosen = void 0;

      if (_.isObject(config) && _.isArray(config.exclude)) {
        excludeList = config.exclude;
      }

      // !keyChosen -- it's the first iteration
      // _.includes(excludeList, keyChosen) -- key already chosen, but should be excluded
      while (!keyChosen || _.includes(excludeList, keyChosen)) {
        keyChosen = this.chooseKeyFromProportions();
      }
      return keyChosen;
    }
  }, {
    key: 'roll',
    value: function roll() /* value or config */{
      // optional
      var keyFromArgs = arguments[0];
      var keyIsAvailable = _.has(this.originalProportions, keyFromArgs);
      var keyChosen = void 0;

      if (!_.isUndefined(keyFromArgs) && keyIsAvailable) {
        // this.roll(value) is called
        keyChosen = keyFromArgs;
      } else if (!_.isUndefined(keyFromArgs) && !keyIsAvailable && !_.isObject(keyFromArgs)) {
        // this.roll(value) is called incorectly
        throw new Error('The value ' + keyFromArgs + ' is not in your originalProportions.');
      } else {
        // the normal this.roll()

        keyChosen = this.selectKeyAtRandom(arguments[0]);

        if (!this.passCriteria(keyChosen)) {
          return this.roll();
        }
      }
      this.shiftProportions(keyChosen);
      this.rollSideEffects(keyChosen);
      return keyChosen;
    }
  }, {
    key: 'undo',
    value: function undo() {
      if (!this.keepHistory) {
        // throw if there's no keepHistory
        throw new Error('To undo, set `{keepHistory: true}`.');
      }
      var previousChoice = this.historyArray.pop(); // ! writes to historyArray
      this.historyMapping[previousChoice] = this.historyMapping[previousChoice] - 1;

      this.reverseShiftProportions(previousChoice);

      return previousChoice;
    }
  }, {
    key: 'reset',
    value: function reset() {
      this.proportions = _.cloneDeep(this.originalProportions);
      if (this.keepHistory) {
        this.historyMapping = _.mapValues(this.originalProportions, _.constant(0));
        this.historyArray = [];
      }
    }
  }]);

  return Brng;
}();

Brng.defaultProportions = {
  one6SidedDie: { 1: 1, 2: 1, 3: 1, 4: 1, 5: 1, 6: 1 },
  two6SidedDice: { 2: 1, 3: 2, 4: 3, 5: 4, 6: 5, 7: 6, 8: 5, 9: 4, 10: 3, 11: 2, 12: 1 },
  coin: { heads: 1, tails: 1 }
};
exports.default = Brng;
module.exports = exports['default'];
